"""Copyright 2016-2017 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License."""

from traceutils.record_storage import RecordStorage
import logging


class Oracle:
    """Traces checking engine."""

    def __init__(self, window_size, first_diff_stop):
        """Default constructor."""

        self.window_size = window_size
        self.first_diff_stop = first_diff_stop

        # Mapping from trace names to related record storages.
        self._trace_rec_map = {}

    def register_record(self, record, trace_name):
        """Registers log record of the trace with the specified name in this oracle."""

        logging.debug('Try to register record in trace \'%s\': %s' % (trace_name, record.to_string()))

        if record is None:
            logging.debug('Can\'t parse the record: %s' % record)
            return -1

        traces_with_record = Oracle.appear_in_other_traces(self, record, trace_name)

        if len(traces_with_record):
            # If the record was in other traces, then it should be removed from everywhere
            for rec_trace_name in traces_with_record.keys():
                rec_storage = self._trace_rec_map.get(rec_trace_name)
                repeated_record = traces_with_record[rec_trace_name]
                logging.debug(
                        'Remove record \'%s\' from trace because of coincidence: %s'
                        % (repeated_record.to_string(), rec_trace_name))
                rec_storage.remove_record(repeated_record)

        else:
            if trace_name not in self._trace_rec_map:
                logging.debug('A new storage is created for: %s' % trace_name)
                self._trace_rec_map[trace_name] = RecordStorage(self.window_size)

            # Check whether trace records are ordered by time
            record_storage = self._trace_rec_map[trace_name]
            if not record_storage.empty():

                last_record = record_storage.get_last_record()
                last_record_time = last_record.get_time()

                if last_record_time > record.get_time():
                    logging.error('Can\'t register a record with time %s; the last record has %s'
                                  % (record.get_time(), last_record.get_time()))
                    return -3

                if (record_storage.get_duration() > self.window_size) and self.first_diff_stop:

                    logging.debug(
                            'First difference is detected at %s record, clear the late and exit.'
                            % last_record.to_string())

                    for map_trace_name in self._trace_rec_map.keys():
                        map_storage = self._trace_rec_map[map_trace_name]
                        storage_duration = map_storage.get_duration()
                        if storage_duration > self.window_size:
                            min_skipped_idx = map_storage.get_max_time() - self.window_size
                            logging.debug('The minimal skip index: %d' % min_skipped_idx)
                            map_storage.clear_records_after(min_skipped_idx)
                    return -2

            record_storage.add_record(record)
        return 0

    def appear_in_other_traces(self, record, trace_name):
        """Returns the map of trace names, that contain the record that matches to the specified one."""

        logging.debug('Whether record appears in other traces than \'%s\': %s' % (trace_name, record.to_string()))

        traces_with_record = {}
        for map_trace_name, map_storage in self._trace_rec_map.iteritems():
            match_record = map_storage.match_record(record)
            if (not (match_record is None)) and (map_trace_name != trace_name):
                logging.debug(
                        'The same record has been found in %s trace for: %s' % (map_trace_name, record.to_string()))
                traces_with_record[map_trace_name] = match_record

        return traces_with_record

    def future_record_for_other_traces(self, record, trace_name):
        """Checks whether the specified record from the specified trace is a 'future' one for other traces."""
        for map_trace_name, map_storage in self._trace_rec_map.iteritems():
            if map_trace_name == trace_name:
                continue
            if map_storage.empty():
                continue

            max_time = map_storage.get_max_time()
            record_time = record.get_time()

            return record_time - self.window_size > max_time

        return False

    def check(self):
        """Checks whether traces are equivalent."""
        all_traces_equivalent = True

        for record_storage in self._trace_rec_map.values():
            if not record_storage.empty():
                all_traces_equivalent = False
                break

        if not all_traces_equivalent:
            logging.info('=============================================\n')
            logging.info('WARNING: Non-matched records have been found:\n')
            logging.info('=============================================')
            for trace_name, record_storage in self._trace_rec_map.iteritems():
                if not record_storage.empty():
                    logging.info('\nTrace=%s\n' % trace_name)
                    record_storage.print_all_records()
