"""Copyright 2016-2017 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License."""


from traceutils.trace import Trace
from traceutils.record.instruction_record import InstructionRecord
from traceutils.record.record import Record
from traceutils.record.reg_record import RegisterRecord
from traceutils.record.mem_access_record import MemoryAccessRecord
import logging
import re


class TraceParser:
    """Engine that transforms trace lines into record objects."""

    def __init__(self, file_path):
        """Default constructor."""
        self._trace = Trace(file_path)
        self._last_record = None

    def get_trace(self):
        """Returns the associated trace."""
        return self._trace

    def get_trace_name(self):
        """Returns the name of the current trace."""
        return self._trace.name()

    def get_trace_length(self):
        """Returns the length of this parser trace."""
        return self._trace.get_trace_length()

    def has_next_record(self, skip_equal):
        """Checks whether the engine can provide one more record from the trace."""

        if skip_equal and not (self._last_record is None):

            if not self._trace.has_next_event():
                return False

            while self._trace.has_next_event():

                event_line = self._trace.peek_next_event()
                event_record = self.match_record(event_line)

                if not Record.equal_records(event_record, self._last_record, Record.TIME):
                    return True
                else:
                    self._trace.delete_next_event()
            return False

        return self._trace.has_next_event()

    def next_record(self):
        """Returns record object from the next line of trace."""

        if not self._trace.has_next_event():
            print 'File is empty'
            return None

        event_line = self._trace.get_next_event()
        record = self.match_record(event_line)

        self._last_record = record
        return record

    def get_last_record(self):
        """Returns the record that was previously taken."""
        return self._last_record

    def peek_next_record(self):
        """Returns th current event from the related trace without it's extraction."""

        return self.match_record(self._trace.peek_next_event())

    def match_record(self, event_line):
        """Matches the trace line to one of supported record types."""

        record = None
        recognized = False

        cpu_match = re.match(InstructionRecord.get_format(), event_line)
        if cpu_match:
            recognized = True
            record = InstructionRecord(cpu_match.group(Record.TIME),
                                       cpu_match.group(Record.SCALE),
                                       cpu_match.group(InstructionRecord.CPU),
                                       cpu_match.group(InstructionRecord.COND_CODE),
                                       cpu_match.group(InstructionRecord.INST_ID),
                                       cpu_match.group(InstructionRecord.ADDR),
                                       cpu_match.group(InstructionRecord.OPCODE),
                                       cpu_match.group(InstructionRecord.INST_SET),
                                       cpu_match.group(InstructionRecord.MODE),
                                       cpu_match.group(InstructionRecord.SECURITY),
                                       cpu_match.group(InstructionRecord.DISASM))

        reg_match = re.match(RegisterRecord.get_format(), event_line)
        if reg_match:
            recognized = True
            record = RegisterRecord(reg_match.group(Record.TIME),
                                    reg_match.group(Record.SCALE),
                                    reg_match.group(RegisterRecord.REGISTER),
                                    reg_match.group(RegisterRecord.VALUE))

        mem_match = re.match(MemoryAccessRecord.get_format(), event_line)
        if mem_match:
            recognized = True
            record = MemoryAccessRecord(mem_match.group(Record.TIME),
                                        mem_match.group(Record.SCALE),
                                        mem_match.group(MemoryAccessRecord.RW),
                                        mem_match.group(MemoryAccessRecord.SIZE),
                                        mem_match.group(MemoryAccessRecord.ATTRIB),
                                        mem_match.group(MemoryAccessRecord.ADDR),
                                        mem_match.group(MemoryAccessRecord.DATA))

        if not recognized:
            logging.warning('In trace: %s\n Can\'t parse: %s' % (self.get_trace_name(), event_line))
            return None

        return record

    def dump_rest_lines(self):
        """Dumps to the stdout the rest of this trace's lines."""
        self._trace.dump_rest_lines()
